;;
;; 
;; An extension for specifying Hodgkin-Huxley type dynamics in NEMO
;; systems.
;;
;; Copyright 2008 Ivan Raikov and the Okinawa Institute of Science and Technology
;;
;; This program is free software: you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation, either version 3 of the
;; License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.
;;
;; A full copy of the GPL license can be found at
;; <http://www.gnu.org/licenses/>.
;;

(module nemo-hh

	(nemo:hh-transformer)

	(import scheme chicken data-structures srfi-1 srfi-13)
	
	(require-extension matchable environments nemo-core)

(define (s+ . lst)    (string-concatenate (map ->string lst)))

(define (p$ p n) (string->symbol (s+ (->string p) "_" (->string n))))


(define (lookup-field k lst . rest)
  (let-optionals rest ((default #f))
   (let ((v (alist-ref k lst)))
     (if v (first v) default))))


(define (check-names ion env . names)
  (for-each (lambda (name)
	      (if (environment-includes? env name)
		  (nemo:error 'nemo:hh-transformer "quantity " name " in ionic conductance declaration " ion
			     "is already declared elsewhere")))
	    names))

(define (check-decls ion names alst . rest)
  (let-optionals rest ((raise-exception? #t))
    (if raise-exception?
	(for-each (lambda (name) 
		    (if (not (alist-ref name alst))
			(nemo:error 'nemo:hh-transformer "required quantity " name 
				   " is not present in ionic conductance declaration " ion)))
		  names)
	(every (lambda (name) (alist-ref name alst)) names))))


(define (handle thunk dflt)
  (condition-case (thunk)
		  [(exn) dflt]))

(define (hh-ionic-gate-transform sys parse-expr eval-const env-extend! add-external! component-extend! comp en)
  (define (and-parse-expr x . rest) (and x (apply parse-expr (cons x rest))))
  (match en
	 ((or (('hh 'ionic 'conductance)  ('name (? symbol? ion)) . alst)
	      (('hh-ionic-gate)    ('name (? symbol? ion)) . alst))
	  (check-decls ion '(m-power h-power) alst)
	  (let ((suffix (->string ion))
		(m-power (eval-const sys (parse-expr (lookup-field 'm-power alst) 
						     `(hh-ionic-gate ,ion (m-power))
						     )))
		(h-power (eval-const sys (parse-expr (lookup-field 'h-power alst 0)
						     `(hh-ionic-gate ,ion (h-power))
						     ))))
	    
	    (if (not (and (integer? m-power) (positive? m-power)))
		(nemo:error 'nemo:hh-transformer 
			   "m-power value in ionic conductance declaration " ion
			   " must be a positive integer"))
	    
	    ;; check for required decls in m
	    (check-decls ion '(initial-m) alst)
	    (if (not (check-decls ion '(m-inf m-tau) alst #f))
		(check-decls ion '(m-alpha m-beta) alst))
	    
	    ;; check for required decls in h
	    (if (positive? h-power) 
		(begin (check-decls ion '(initial-h) alst)
		       (if (not (check-decls ion '(h-inf h-tau) alst #f))
			   (check-decls ion '(h-alpha h-beta) alst))))
	    
	    (if (not (and (integer? h-power) (or (zero? h-power) (positive? m-power))))
		(nemo:error 'nemo:hh-transformer 
			   "h-power value in ionic conductance declaration " ion
							" must be a positive integer"))
					 
	    (let* ((initial-m  ((lambda (x) 
				  (let ((expr (parse-expr x `(hh-ionic-gate ,ion (initial-m)))))
				    (handle (lambda () (eval-const sys expr)) expr)))
				(lookup-field 'initial-m alst)))
		   (m-inf      (and-parse-expr (lookup-field 'm-inf alst)
					       `(hh-ionic-gate ,ion (m-inf))))
		   (m-tau      (and-parse-expr (lookup-field 'm-tau alst)
					       `(hh-ionic-gate ,ion (m-tau))))
		   (m-inf-sym  (p$ ion 'm-inf))
		   (m-tau-sym  (p$ ion 'm-tau))
		   (m-alpha    (or (and-parse-expr (lookup-field 'm-alpha alst)
						   `(hh-ionic-gate ,ion (m-alpha)))
				   `(/ ,m-inf-sym ,m-tau-sym)))
		   (m-beta     (or (and-parse-expr (lookup-field 'm-beta alst)
						   `(hh-ionic-gate ,ion (m-beta)))   
				   `(/ (- 1 ,m-inf-sym) ,m-tau-sym)))
		   (open       'O)
		   (closed     'C)
		   (mst        `((power ,m-power)  (open  ,open)
				 (transitions (<-> ,closed ,open ,m-alpha ,m-beta)))))

	      (if m-inf (env-extend! m-inf-sym '(asgn) 'none `(rhs ,m-inf)))
	      (if m-tau (env-extend! m-tau-sym '(asgn) 'none `(rhs ,m-tau)))
	      (apply env-extend! (cons* (p$ ion 'm) '(reaction) initial-m mst))
	      (add-external! (p$ ion 'm) 'output)
	      (component-extend! comp (p$ ion 'm))
	      (if m-inf (component-extend! comp m-inf-sym))
	      (if m-tau (component-extend! comp m-tau-sym))
	      )
	    
	    (if (positive? h-power)
		(let* ((initial-h  ((lambda (x)
				      (let ((expr (parse-expr x `(hh-ionic-gate ,ion (initial-h)))))
					(handle (lambda () (eval-const sys expr)) expr)))
				    (lookup-field 'initial-h alst)))
		       (h-inf      (and-parse-expr (lookup-field 'h-inf alst)
						   `(hh-ionic-gate ,ion (h-inf))))
		       (h-tau      (and-parse-expr (lookup-field 'h-tau alst)
						   `(hh-ionic-gate ,ion (h-tau))))
		       (h-alpha    (or (and-parse-expr (lookup-field 'h-alpha alst)
						       `(hh-ionic-gate ,ion (h-alpha)))
				       `(/ ,h-inf ,h-tau)))
		       (h-beta     (or (and-parse-expr (lookup-field 'h-beta alst)
						       `(hh-ionic-gate ,ion (h-beta)))
				       `(/ (- 1 ,h-inf) ,h-tau)))

		       (open       'O)
		       (closed     'C)
		       (hst        `((power ,h-power)
				     (open  ,open)
				     (transitions (<-> ,closed ,open ,h-alpha ,h-beta) ))))

		  (apply env-extend! (cons* (p$ ion 'h) '(reaction) initial-h hst))
		  (add-external! (p$ ion 'h) 'output)
		  (component-extend! comp (p$ ion 'h))
		  ))))
	 (else (list))))

(define (nemo:hh-transformer sys . rest)
  (let-optionals rest ((parse-expr (lambda (x . rest) (identity x))))
   (let ((new-sys  (nemo:env-copy sys)))
     (match-let ((($ nemo:quantity 'DISPATCH  dis) (environment-ref new-sys (nemo-intern 'dispatch))))
      (let* ((eval-const         (dis 'eval-const))
	     (env-extend!        ((dis 'env-extend!) new-sys))
	     (add-external!      ((dis 'add-external!) new-sys))
	     (component-extend!  ((dis 'component-extend!) new-sys))
	     (indent  0)
	     (indent+ (+ 2 indent )))
	(let recur ((comp-name (nemo-intern 'toplevel)))
	  (let* ((comp-symbols   ((dis 'component-symbols) new-sys comp-name))
		 (subcomps       ((dis 'component-subcomps) new-sys comp-name)))
	    (for-each (lambda (sym)
			(hh-ionic-gate-transform  
			 new-sys parse-expr (dis 'eval-const) env-extend! add-external! component-extend!
			 comp-name (environment-ref new-sys sym)))
		      comp-symbols)
	    (for-each recur (map third subcomps))))
	new-sys)))))
)